/* crlcache.c - LDAP access
 *      Copyright (C) 2002 Klarlvdalens Datakonsult AB
 *      Copyright (C) 2003 g10 Code GmbH
 *
 * This file is part of DirMngr.
 *
 * DirMngr is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * DirMngr is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <assert.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>

#include "dirmngr.h"
#include "crlcache.h"
#include "crlfetch.h"
#include "misc.h"


/* Layout of the Cache DB file:

   We use records of variable length with this structure

   n  bytes  Serialnumber (binary) used as key
             thus there is no need to store the length explicitly with DB2.
   1  bytes  Reason for revocation (currently the KSBA reason flags are used)
   15 bytes  ISO date of revocation (e.g. 19980815T142000)
             Note that there is no terminating 0 stored.
*/

/* Definition of one cached item. */
struct crl_cache_entry_s 
{
  struct crl_cache_entry_s *next;
  char  *url;
  char  *issuer;
  char  *issuer_hash;
  ksba_isotime_t this_update;
  ksba_isotime_t next_update;
  DB* dbf;
};


/* Definition of the entire cache object. */
struct crl_cache_s 
{
  crl_cache_entry_t entries;
};


/* This structure is used internally to build a list of records to be
   inserted into the DB. */
struct record_list_s {
  struct record_list_s *next;
  unsigned char data[1+15];
  size_t keylen;
  unsigned char key[1];
};



/*

   Wrappers for DB access. 

*/
static int 
mydbopen (const char* file, DB** dbpp)
{
#ifdef HAVE_DB4_DB_H
  int rc = db_create( dbpp, 0, 0 );
  if( !rc ) {
    rc = (*dbpp)->open( *dbpp, file, 0, DB_HASH, DB_CREATE, S_IRUSR|S_IWUSR);
  }
  return rc;
#elif HAVE_DB2_DB_H
  return db_open( file, DB_HASH, DB_CREATE, S_IRUSR|S_IWUSR, NULL, NULL, dbpp );
#endif
}

static int 
mydbclose( DB* dbp )
{
  return dbp->close( dbp, 0 );
}

static int 
mydbget( DB* dbp, DBT *key, DBT *data )
{
  return dbp->get( dbp, NULL, key, data, 0 );
}

static int 
mydbput( DB* dbp, DBT *key, DBT *data )
{
  return dbp->put( dbp, NULL, key, data, 0 );
}





/* 
   Create ae directory if it does not yet exists.  Returns on
   success, or -1 on error.
 */
static int
create_directory_if_needed (const char *name)
{
  DIR *dir;
  char *fname;

  fname = make_filename (opt.homedir, name, NULL);
  dir = opendir (fname);
  if (!dir)
    {
      log_info ("creating directory `%s'\n", fname);
      if (mkdir (fname, S_IRUSR|S_IWUSR|S_IXUSR) )
        {
          log_error ("error creating directory `%s': %s\n",
                     fname, strerror (errno));
          xfree (fname);
          return -1;
        }
    } 
  else
    closedir (dir);
  xfree (fname);
  return 0;
}

/* Remove all files from the cache directory.  If FORCE is not true,
   some sanity checks on the filenames are done. Return 0 if
   everything went fine. */
static int
cleanup_cache_dir (int force)
{
  char *dname = make_filename (opt.homedir, DBDIR, NULL);
  DIR *dir;
  struct dirent *de;
  int problem = 0;

  if (!force)
    { /* Very minor sanity checks. */
      if (!strcmp (dname, "~/") || !strcmp (dname, "/" ))
        {
          log_error ("ignoring database dir `%s'\n", dname);
          xfree (dname);
          return -1;
        }
    }

  dir = opendir (dname);
  if (!dir)
    {
      log_error ("error reading directory `%s': %s\n", dname, strerror (errno));
      xfree (dname);
      return -1;
    }

  while ((de = readdir (dir)))
    {
      if (strcmp (de->d_name, "." ) && strcmp (de->d_name, ".."))
        {
          char *dbfname = make_filename (dname, de->d_name, NULL);
          int okay;
          struct stat sbuf;
     
          if (force)
            okay = 1;
          else
            okay = (!stat (dbfname, &sbuf) && S_ISREG (sbuf.st_mode));

          if (okay)
            {
              log_info ("removing cache file `%s'\n", dbfname);
              if (unlink (dbfname))
                {
                  log_error ("failed to remove `%s': %s\n",
                             dbfname, strerror (errno));
                  problem = -1;
                }
            }
          else
            log_info ("not removing file `%s'\n", dbfname);
          xfree (dbfname);
        }
    }	  
  xfree (dname);
  closedir (dir);
  return problem;
}


/* Read the next line from the file FP and return the line in an
   malloced buffer.  Return NULL on error or EOF.  The length of the
   line is limited to 1024 characters.  The trailing linefeed has been
   removed, the fucntion will read the last line of a file, even if
   that is not terminated by a LF. */
static char *
next_line_from_file (FILE *fp)
{
  char buf[1024];
  int len;

  if (!fgets (buf, sizeof buf, fp))
    return NULL;

  len = strlen (buf);
  if (len && buf[len-1] == '\n')
    buf[--len] = 0;
  return xstrdup (buf);
}


/* Release the cache object. */
static void
release_cache (crl_cache_t cache)
{
  crl_cache_entry_t entry, entry2;

  if (!cache)
    return;

  for (entry = cache->entries; entry; entry = entry2)
    {
      entry2 = entry->next;
      xfree (entry->issuer_hash);
      if (entry->dbf)
        mydbclose (entry->dbf);
      xfree (entry);
  }
  cache->entries = NULL;
  xfree (cache);
}


/* Create a new cache object and return it in R_CACHE. */
static gpg_error_t
crl_cache_new (crl_cache_t *r_cache)
{
  crl_cache_t cache = xcalloc (1, sizeof *cache);
  char *fname;
  FILE *contents;
  char *line;
  int err = 0;

  fname = make_filename (opt.homedir, DBCONTENTSFILE, NULL);
  contents = fopen( fname, "r" );
  if (!contents)
    {
      log_error ("can't open DB contents file `%s': %s",
                 fname, strerror (errno));
      err = gpg_error (GPG_ERR_CONFIGURATION);
      goto leave;
    }

  /* Make sure that the directory exists, try to create if otherwise.
     We don't bail out on error here, since the actual open/create
     functions will do so anyway. */
  if (!create_directory_if_needed (NULL))
    create_directory_if_needed (DBDIR);

  /* Read file version number. */
  line = next_line_from_file (contents);
  if (line)
    {
      if (strcmp (line, DBCONTENTSVERSION))
        {
          /* Not same version, don't read file.  We better now remove
             all files from that directory - they are not longer useful.
             Yeah, it's a cache dude. */
          xfree (line); line = NULL;
          cleanup_cache_dir (1);
          err = gpg_error (GPG_ERR_CONFIGURATION);
          goto leave;
        }
      xfree (line); line = NULL;
    }

  /* Walk over all entries from the contents file. */
  for (;;)
    {
      crl_cache_entry_t entry = xcalloc (1, sizeof *entry );
      int rc;
      char *dbfname;

      entry->url = line = next_line_from_file (contents);
      if (line)
        entry->issuer = line = next_line_from_file (contents);
      if (line)
        entry->issuer_hash = line = next_line_from_file (contents);
      if (line)
        line = next_line_from_file (contents);
      if (line)
        copy_time (entry->this_update,line);
      if (line)
        line = next_line_from_file (contents);
      if (line)
        copy_time (entry->next_update,line);

      if (!line)
        {
          if (ferror (contents))
            log_error ("error reading `%s': %s\n", fname, strerror (errno));
          else if (entry->url)
            log_error ("file `%s' has an invalid structure\n", fname);
          else
            {
              xfree (entry);
              err = 0;
              break; /* Ready. */
            }
          xfree (entry);
          err = gpg_error (GPG_ERR_CONFIGURATION);
          goto leave;
        }

      dbfname = make_filename (opt.homedir, DBDIR,
                               entry->issuer_hash, NULL);
      if (opt.verbose)
        log_info ("opening DB file `%s'\n", dbfname );

      rc = mydbopen (dbfname, &entry->dbf);
      if (rc)
        {
          log_error ("error opening DB file `%s': %s (rc=%d) - skipped",
                     dbfname, strerror(errno), rc );
          xfree (entry);
        }
      else
        {
          entry->next = cache->entries;
          cache->entries = entry;
        }
      xfree (dbfname);
    }


 leave:
  if (contents)
    fclose (contents);
  xfree (fname);
  if (err)
    {
      release_cache (cache);
      cache = NULL;
    }
  *r_cache = cache;
  return err;
}


/* Find ISSUER_HASH in our cache FIRST. This may be used to enumerate
   the linked list we use to keep the CRLs of an issuer. */

static crl_cache_entry_t 
find_entry (crl_cache_entry_t first, const char *issuer_hash)
{
  while (first && strcmp (issuer_hash, first->issuer_hash) )
    first = first->next;

  if (DBG_LOOKUP)
    {
      if (first)
        log_debug ("find_entry found entry %p (issuer=%s)\n",
                   first, first->issuer_hash );
      else 
        log_debug ("find_entry found no entry\n");
    }
  return first;    
}


static void 
save_contents (crl_cache_entry_t entry)
{
  char *fname = make_filename (opt.homedir, DBCONTENTSFILE, NULL);
  char *fnametmp = make_filename (opt.homedir, DBCONTENTSFILE ".tmp", NULL );
  FILE *contents;

  contents = fopen (fnametmp, "w");
  if (!contents)
    {
      log_error ("error writing contents file `%s': %s",
                  fnametmp, strerror(errno) );
      goto failure;
    }
  
  /* Write out the file. */
  fprintf (contents, "%s\n", DBCONTENTSVERSION);
  for (; entry; entry = entry->next)
    {
      fprintf (contents, "%s\n", entry->url);
      fprintf (contents, "%s\n", entry->issuer);
      fprintf (contents, "%s\n", entry->issuer_hash);
      fprintf (contents, "%s\n", entry->this_update);
      fprintf (contents, "%s\n", entry->next_update);
    }

  if (fclose (contents))
    {
      log_error ("error closing `%s':%s\n", fnametmp, strerror (errno));
      remove (fnametmp);
      goto failure;
    }      
  
  if (rename( fnametmp, fname ))
    {
      log_error ("error updating contents file `%s': %s",
                 fname, strerror (errno));
      remove (fnametmp);
      goto failure;
    }
  xfree (fname);
  xfree (fnametmp);
  return;

 failure:
  /* fixme: we should fail more gracefully; i.e. return an error
     code. */
  xfree (fname);
  xfree (fnametmp);
  log_error ("terminating\n");
  dirmngr_exit (2);
}


/* Create a new CRL cache. Returns the handle to this cache. Will
   never fail. */
crl_cache_t
crl_cache_init(void)
{
  crl_cache_t cache;
  gpg_error_t err;

  err = crl_cache_new (&cache);
  if (err)
    log_fatal ("failed to create a new cache object: %s\n", gpg_strerror (err));
  return cache;
}


/* Remove the cache information and all its resources.  Note that we
   still keep the cache on disk. */
void 
crl_cache_deinit (crl_cache_t cache)
{
  if (!cache)
    return;

  save_contents (cache->entries);
  release_cache (cache);
}


/* Delete the cache from disk. Return 0 on success.*/
int 
crl_cache_flush (void)
{
  int problem = 0;
  char *fname = make_filename (opt.homedir, DBCONTENTSFILE, NULL);

  if (unlink (fname))
    {
      log_error ("error removing database content file `%s': %s\n",
                 fname, strerror (errno) );
      problem = -1;
      /* Try to continue anyway. */
    }
  xfree (fname);

  if (cleanup_cache_dir (0))
    problem = -1;
  return problem;
}


/* Check whether the certificate identified by ISSUER_HASH and
   SERIALNO is valid; i.e. not listed in our cache. */
crl_cache_result_t 
crl_cache_isvalid (crl_cache_t cache,
                   const char *issuer_hash, const char *serialno)
{
  crl_cache_result_t retval;
  DBT key;
  DBT result;
  int rc;
  crl_cache_entry_t entry;
  dirmngr_isotime_t current_time;
  unsigned char snbuf_buffer[50];
  unsigned char *snbuf;
  size_t n;

  entry = find_entry (cache->entries, issuer_hash);
  if (!entry)
    {
      log_info ("no CRL available for issuer `%s'\n", issuer_hash );
      return CRL_CACHE_DONTKNOW;
    }

  get_isotime (current_time);
  if (strcmp (entry->next_update, current_time) < 0 )
    {
      log_info ("CRL for issuer (id=%s) too old, we need to update\n",
                issuer_hash);
      return CRL_CACHE_DONTKNOW;
    }
  
  memset (&key, 0, sizeof key );
  memset (&result, 0, sizeof result );

  n = strlen (serialno)/2+1;
  if (n < sizeof snbuf_buffer - 1)
    snbuf = snbuf_buffer;
  else
    snbuf = xmalloc (n);

  key.data = snbuf;
  key.size = unhexify (snbuf, serialno);

  rc = mydbget (entry->dbf, &key, &result);
  if (!rc) 
    {
      int reason = result.size? *((unsigned char*)result.data) : 0;

      if (opt.verbose)
        log_info ("S/N %s is not valid; reason=%02X  date=%.15s\n",
                  serialno, reason, (char*)result.data+1);
      retval = CRL_CACHE_INVALID;
    }
  else if (rc == DB_NOTFOUND)
    {
      if (opt.verbose)
        log_info ("S/N %s is valid, it is not listed in the CRL\n", serialno );
      retval = CRL_CACHE_VALID;
    }
  else 
    {
      log_error ("error getting data from DB (rc=%d): %s\n",
                 rc, strerror(errno) );
      retval = CRL_CACHE_DONTKNOW;
    }

  if (snbuf != snbuf_buffer)
    xfree (snbuf);

  return retval;
}



/* Return the certificate for ISSUER or NULL if it can't be found. */
static ksba_cert_t 
get_issuer_cert (ctrl_t ctrl, const char *issuer)
{
  gpg_error_t err;
  ksba_cert_t issuer_cert = NULL;
  unsigned char *cvalue;
  size_t cvaluelen;

  /* Ask back to the service requester to return the certificate.
     This is because we can assume that he already used the
     certificate while checking for the CRL. */
  issuer_cert = get_issuer_cert_local (ctrl, issuer);
  if (issuer_cert)
    return issuer_cert; /* Done. */

  /* If this fails do a regular certificate lookup. */
  if (DBG_LOOKUP)
    log_debug ("certificate for CRL issuer not returned by caller"
               " - doing lookup\n");

  err = ca_cert_fetch (issuer, &cvalue, &cvaluelen );
  if (err)
    {
      log_error ("error fetching certificate for issuer: %s\n",
                 gpg_strerror (err));
      return NULL;
    }

  err = ksba_cert_new (&issuer_cert);
  if (!err)
    {
      err = ksba_cert_init_from_mem (issuer_cert, cvalue, cvaluelen);
      if (err)
        {
          ksba_cert_release (issuer_cert);
          issuer_cert = NULL;
        }
    }
  if (err)
    log_error ("invalid issuer certificate: %s\n", gpg_strerror (err) );

  xfree( cvalue );

  return issuer_cert;
}


/* Create a new entry and DB file for ISSUER and return it.  Returns
   NULL on error. */
static crl_cache_entry_t 
db_entry_new (const char *url, const char* issuer, const char* issuer_hash, 
	      const dirmngr_isotime_t thisupdate,
              const dirmngr_isotime_t nextupdate )
{
  char *fname;
  crl_cache_entry_t entry;
  int rc;

  entry = xcalloc ( 1, sizeof *entry );

  entry->url         = xstrdup (url);  
  entry->issuer      = xstrdup (issuer);
  entry->issuer_hash = xstrdup (issuer_hash);
  copy_time (entry->this_update, thisupdate);
  copy_time (entry->next_update, nextupdate);

  fname = make_filename (opt.homedir, DBDIR, entry->issuer_hash, NULL);
  rc = mydbopen (fname, &entry->dbf);
  if(rc)
    log_error ("error opening DB (rc=%d): %s\n", rc, strerror (errno));
  xfree (fname);  
  if(rc)
    {
      xfree(entry);
      entry = NULL;
    }
  return entry;
}


/* Prepare a hash context for the signature verification.  Input is
   the CRL and the output is the hash context MD as well as the uses
   algorithm identifier ALGO. */
static gpg_error_t
start_sig_check (ksba_crl_t crl, gcry_md_hd_t *md, int *algo) 
{
  gpg_error_t err;
  const char *algoid;

  algoid = ksba_crl_get_digest_algo (crl);
  *algo = gcry_md_map_name (algoid);
  if (!*algo) 
    {
      log_error ("unknown hash algorithm `%s'\n", algoid? algoid:"?");
      return gpg_error (GPG_ERR_DIGEST_ALGO);
    }

  err = gcry_md_open (md, *algo, 0);
  if (err)
    {
      log_error ("md_open for algorithm %d failed: %s\n",
                 *algo, gcry_strerror (err));
      return err;
    }
  
  ksba_crl_set_hash_function (crl, HASH_FNC, *md);  
  return 0;
}


/* Finish a hash context and verify the signature.  This function
   should return 0 on a good signature, GPG_ERR_BAD_SIGNATURE if the
   signature does not verify or any other error code. CRL is the CRL
   object we are working on, MD the hash context and ISSUER_CERT the
   certificate of the CRL issuer. */
static gpg_error_t
finish_sig_check (ksba_crl_t crl, gcry_md_hd_t md, int algo,
                  ksba_cert_t issuer_cert)
{
  gpg_error_t err;
  ksba_sexp_t sigval = NULL, pubkey = NULL;
  const char *s;
  char algoname[50];
  size_t n;
  gcry_sexp_t s_sig = NULL, s_hash = NULL, s_pkey = NULL;
  int i;

  gcry_md_final (md);

  /* Get and convert the signature value. */
  sigval = ksba_crl_get_sig_val (crl);
  n = gcry_sexp_canon_len (sigval, 0, NULL, NULL);
  if (!n) 
    {
      log_error ("got an invalid S-expression from libksba\n");
      err = gpg_error (GPG_ERR_INV_SEXP);
      goto leave;
    }
  err = gcry_sexp_sscan (&s_sig, NULL, sigval, n);
  if (err) 
    {
      log_error ("converting S-expression failed: %s\n", gcry_strerror (err));
      goto leave;
    }
	
  /* Get and convert the public key for the issuer certificate. */
  pubkey = ksba_cert_get_public_key (issuer_cert);
  n = gcry_sexp_canon_len (pubkey, 0, NULL, NULL);
  if (!n) 
    {
      log_error ("got an invalid S-expression from libksba\n");
      err = gpg_error (GPG_ERR_INV_SEXP);
      goto leave;
    }
  err = gcry_sexp_sscan (&s_pkey, NULL, pubkey, n);
  if (err) 
    {
      log_error ("converting S-expression failed: %s\n", gcry_strerror (err));
      goto leave;
    }

  /* Create an S-expression with the actual hash value. */
  s = gcry_md_algo_name (algo);
  for (i=0; *s && i < sizeof(algoname) - 1; s++, i++)
    algoname[i] = ascii_tolower (*s);
  algoname[i] = 0;
  err = gcry_sexp_build (&s_hash, NULL, "(data(flags pkcs1)(hash %s %b))", 
                         algoname,
                         gcry_md_get_algo_dlen (algo), gcry_md_read (md, algo));
  if (err) 
    {
      log_error ("creating S-expression failed: %s\n", gcry_strerror (err));
      goto leave;
    }

  /* Pass this on to the signature verification. */
  err = gcry_pk_verify (s_sig, s_hash, s_pkey);

 leave:
  xfree (sigval);
  xfree (pubkey);
  gcry_sexp_release (s_sig);
  gcry_sexp_release (s_hash);
  gcry_sexp_release (s_pkey);
  
  return err;
}

/* Workhorse of the CRL loading machinery. */ 
static int 
crl_parse_insert (ctrl_t ctrl, crl_cache_t cache,
                  const char* url, ksba_crl_t crl )
{  
  gpg_error_t err;
  ksba_stop_reason_t stopreason;
  crl_cache_entry_t entry = NULL;
  int new_entry = 0;
  ksba_cert_t issuer_cert = NULL;
  gcry_md_hd_t md = NULL;
  int algo = 0;
  size_t n;
  struct record_list_s *record_list = NULL;
  struct record_list_s *rl, *rl2;
  char *issuer = NULL;
  char *issuer_hash = NULL;
  ksba_isotime_t thisupdate; 
  ksba_isotime_t nextupdate; 
  dirmngr_isotime_t current_time;

  /* Start of the KSBA parser loop. */
  do
    {
      err = ksba_crl_parse (crl, &stopreason);
      if (err)
        {
          log_error ("ksba_crl_parse failed: %s\n", gpg_strerror (err) );
          goto failure;
        }

      switch (stopreason)
        {
        case KSBA_SR_BEGIN_ITEMS:
          {
            if (start_sig_check (crl, &md, &algo ))
              goto failure;
          
            err = ksba_crl_get_issuer (crl, &issuer);
            if( err )
              {
                log_error ("no issuer found in CRL: %s\n", gpg_strerror (err) );
                err = gpg_error (GPG_ERR_INV_CRL);
                goto failure;
              }
        
            issuer_cert = get_issuer_cert (ctrl, issuer);
            if (!issuer_cert)
              {
                err = gpg_error (GPG_ERR_MISSING_CERT);
                goto failure;
              }
        
            /* Create an hex encoded SHA-1 hash of the issuer DN to be
               used as the key for the cache. */
            issuer_hash = hashify_data (issuer, strlen (issuer));

            new_entry = 0;
            entry = find_entry (cache->entries, issuer_hash);

            err = ksba_crl_get_update_times (crl, thisupdate, nextupdate);
            if (err)
              {
                log_error ("error getting update times of CRL: %s\n",
                           gpg_strerror (err));	  
                err = gpg_error (GPG_ERR_INV_CRL);
                goto failure;
              }

            if (opt.verbose)
              log_info ("update times of this CRL: this=%s next=%s\n", 
                        thisupdate, nextupdate);   

            if (!entry) 
              {
                if (opt.verbose)
                  log_info ("creating new entry for this issuer\n");
                entry = db_entry_new (url, issuer, issuer_hash,
                                      thisupdate, nextupdate);
                if (!entry)
                  {
                    err = gpg_error (GPG_ERR_GENERAL);
                    goto failure;
                  }
                new_entry = 1;
              }
            else
              {
                if (opt.verbose)
                  log_info ("updating entry for this issuer\n");
                copy_time (entry->this_update, thisupdate);
                copy_time (entry->next_update, nextupdate);
              }
        
            ksba_free(issuer);
            issuer = NULL;
          }
          break;
      
        case KSBA_SR_GOT_ITEM:
          {
            ksba_sexp_t serial;
            const unsigned char *p;
            ksba_isotime_t rdate;
            ksba_crl_reason_t reason;

            err = ksba_crl_get_item (crl, &serial, rdate, &reason);
            if (err)
              {
                log_error ("error getting CRL item: %s\n", gpg_strerror (err));
                err = gpg_error (GPG_ERR_INV_CRL);
                ksba_free (serial);
                goto failure;
              }
            p = serial_to_buffer (serial, &n);
            if (!p)
              BUG ();
            rl = xmalloc (sizeof *rl + n - 1);
            memcpy (rl->key, p, n);
            rl->keylen = n;
            rl->data[0] = (reason & 0xff);
            memcpy (rl->data+1, rdate, 15); 
            rl->next = record_list;
            record_list = rl;

            ksba_free (serial);
          }
          break;
      
        case KSBA_SR_END_ITEMS:
          break;
      
        case KSBA_SR_READY:
          {
            err = finish_sig_check (crl, md, algo, issuer_cert);
            if (err)
              {
                log_error ("CRL signature verification failed: %s\n", 
                           gpg_strerror (err));
                goto failure;
              }
          }
          break;
      
        default:
          log_debug( "unknown stop reason\n");
          err = gpg_error (GPG_ERR_BUG);
          goto failure;
        }
    } 
  while (stopreason != KSBA_SR_READY);

  assert (!err);
  assert (entry);
  /* At this point we have checked the signature of the CRL and found
     that everything is fine.  Now its time to actually update the
     on-disk cache. */
  for (rl = record_list; rl ; rl = rl->next)
    {
      DBT key, val;   
      int retcode;
	
      memset (&key, 0, sizeof key );
      memset (&val, 0, sizeof val );

      key.data = rl->key;
      key.size = rl->keylen;
 
      val.data = rl->data;
      val.size = 1 + 15;
	
      retcode = mydbput( entry->dbf, &key, &val );
      if (retcode)
        {
          log_error ("error inserting item into DB: DB rc=%d\n", retcode);
          err = gpg_error (GPG_ERR_GENERAL); /* fixme: can we take
                                                errno here ? */
          goto failure;
        }
    }
  
  /* Make sure we keep an entry for that issuer in our cache structure. */
  if (!find_entry (cache->entries, entry->issuer_hash) )
    {
      entry->next = cache->entries;
      cache->entries = entry;
      new_entry = 0;
    }
  save_contents (cache->entries);

  /* Check whether that new CRL is still not expired. */
  get_isotime (current_time);
  if (strcmp (entry->next_update, current_time) < 0 )
    {
      log_error ("new CRL still too old, it expired on %s",
                 entry->next_update);
      err = gpg_error (GPG_ERR_CRL_TOO_OLD);
    }


 failure:
  ksba_cert_release (issuer_cert);
  for (rl = record_list; rl ; rl = rl2)
    {
      rl2 = rl->next;
      xfree (rl);
    }

  if (entry && new_entry)
    { 
      if (entry->dbf)
        mydbclose( entry->dbf );
      xfree( entry->url );
      xfree( entry->issuer );
      xfree( entry->issuer_hash );
      xfree( entry );
    }
  xfree (issuer_hash);
  return err;
}

/* Insert the CRL retrieved using URL into the cache specified by
   CACHE.  the CRL itself is held in the buffer VALUE of the length
   VALUELEN. */
gpg_error_t
crl_cache_insert (ctrl_t ctrl, crl_cache_t cache, const char *url,
                  const unsigned char *value, size_t valuelen)
{
  gpg_error_t err;
  ksba_crl_t crl;
  ksba_reader_t reader;

  err = ksba_crl_new (&crl);
  if (err)
    {
      log_error ("ksba_crl_new failed: %s\n", gpg_strerror (err));
      goto leave;
    }
  
  err = ksba_reader_new (&reader);
  if (!err)
    err = ksba_reader_set_mem (reader, value, valuelen);
  if (err)
    {
      log_error ("ksba_reader_set_mem failed: %s\n", gpg_strerror (err));
      goto leave;
    }
  err = ksba_crl_set_reader (crl, reader);
  if ( err )
    {
      log_error ("ksba_crl_set_reader failed: %s\n", gpg_strerror (err));
      goto leave;
    }

  err = crl_parse_insert (ctrl, cache, url, crl);

 leave:
  ksba_reader_release (reader);
  ksba_crl_release (crl);
  return err;
}


/* Print one cached entry E in a human readable format to stream
   FP. Return 0 on success. */
static gpg_error_t
list_one_crl_entry (crl_cache_entry_t e, FILE *fp)
{
  DBT key;
  DBT val;
  DBC* cursorp;
  int i, rc;
  
  fputs ("--------------------------------------------------------\n", fp );
  fprintf (fp, "Info about CRL from %s\n", e->url );
  fprintf (fp, " Issuer:\t%s\n", e->issuer );
  fprintf (fp, " Issuer Hash:\t%s\n", e->issuer_hash );
  fprintf (fp, " Valid from %s to %s\n\n", e->this_update, e->next_update); 

#if defined(HAVE_DB4_DB_H) || defined(DB2_CURSOR_4_PARAMS)
#define X ,0
#endif
  rc = e->dbf->cursor( e->dbf, NULL, &cursorp X);
#undef X
  if (rc)
    {
      log_error ("error reading cache entry from db: db cursor rc=%d\n", rc);
      return rc? gpg_error (GPG_ERR_GENERAL) : 0;
    }

  memset (&key, 0, sizeof key );
  memset (&val, 0, sizeof val );

  rc = cursorp->c_get( cursorp, &key, &val, DB_FIRST);
  if (rc)
    {
      log_error ("error reading cache entry from db: db_first rc=%d\n", rc);
      return rc? gpg_error (GPG_ERR_GENERAL) : 0;
    }

  do {
    int reason = val.size? *((unsigned char*)val.data) : 0;
    int any = 0;
    
    fputs ("  ", fp);
    for (i=0; i < key.size; i++)
      fprintf (fp, "%02X", ((unsigned char *)key.data)[i]);
    fputs (":\t reasons( ", fp);
    
    if (reason & KSBA_CRLREASON_UNSPECIFIED)
      fputs( "unspecified ", fp ), any = 1;
    if (reason & KSBA_CRLREASON_KEY_COMPROMISE )
      fputs( "key_compromise ", fp ), any = 1; 
    if (reason & KSBA_CRLREASON_CA_COMPROMISE )
      fputs( "ca_compromise ", fp ), any = 1; 
    if (reason & KSBA_CRLREASON_AFFILIATION_CHANGED )
      fputs( "affiliation_changed ", fp ), any = 1; 
    if (reason & KSBA_CRLREASON_SUPERSEDED )
      fputs( "superseeded", fp ), any = 1; 
    if (reason & KSBA_CRLREASON_CESSATION_OF_OPERATION )
      fputs( "cessation_of_operation", fp ), any = 1; 
    if (reason & KSBA_CRLREASON_CERTIFICATE_HOLD )
      fputs( "certificate_hold", fp ), any = 1; 
    if (reason && !any)
      fputs( "other", fp ); 

    fprintf (fp, ") rdate: %.15s\n", val.size >= 16? ((char*)val.data+1):"??" );
  } 
  while ( !(rc = cursorp->c_get( cursorp, &key, &val, DB_NEXT)));
  
  if (rc == DB_NOTFOUND)
    rc = 0;

  if (rc)
    log_error ("error reading cache entry from db: db_next rc=%d\n", rc);

  return rc? gpg_error (GPG_ERR_GENERAL) : 0;
}


/* Print the contents of the CRL CACHE in a human readable format to
   stream FP. */
gpg_error_t 
crl_cache_list (crl_cache_t cache, FILE *fp) 
{
  crl_cache_entry_t entry;
  gpg_error_t err = 0;

  for (entry = cache->entries; entry && !err; entry = entry->next ) 
    err = list_one_crl_entry (entry, fp);

  return err;
}


/* Load the CRL contain the the file named FILENAME into our cache
   CACHE */
gpg_error_t
crl_cache_load (ctrl_t ctrl, crl_cache_t cache, const char *filename)
{
  gpg_error_t err;
  struct stat sbuf;
  size_t size;
  char* buf;
  FILE* fp;

  if (stat( filename, &sbuf ))
    {
      err = gpg_error_from_errno (errno);
      log_error ("can't stat `%s': %s\n", filename, strerror (errno));
      return err;
    }
  
  size = sbuf.st_size;
  buf = xmalloc (size);
  
  fp = fopen (filename, "r");
  if (!fp)
    {
      err = gpg_error_from_errno (errno);
      log_error ("can't open `%s': %s\n", filename, strerror (errno));
      xfree (buf);
      return err;
  }

  if (fread (buf, size, 1, fp ) != 1)
    {
      if (ferror (fp))
        log_error ("error reading `%s': %s\n", filename, strerror (errno));
      else
        log_error ("premature eof while reading `%s'\n", filename);
      xfree(buf);
      return gpg_error (GPG_ERR_GENERAL);
  }

  err = crl_cache_insert (ctrl, cache, filename, buf, size);
  xfree(buf);
  return err;
}

